# Copyright: Tian Chen and Michael Monagan, 2022

# This code generates benchmark 2 in CASC '21 post conference paper
# It computes the factors of the determinant of a matrix AA with multivariate polynomial entries 
# represented by a black box (monic case only).
# The matrices AA's are generated in Tests.zip. 
# E.g. AA = P2.AA0.P1, where P1 and P2 are upper and lower triangular matrices

CCodeDir := "/home/tca71/Documents/Research/Code/CMSHL_Maple+C/CMSHL/CMSHL3/";

VSolve := define_external('VandermondeSolve64s',
  mm::ARRAY(1..nn,datatype=integer[8]), 
  yy::ARRAY(1..nn,datatype=integer[8]), 
  nn::integer[4],
  aa::ARRAY(1..nn,datatype=integer[8]),
  MM::ARRAY(0..nn,datatype=integer[8]),
  qq::ARRAY(0..nn,datatype=integer[8]), 
  shift::integer[4], p::integer[8], 
  LIB=cat(CCodeDir, "VSolve4.so") ):

Det64s:= define_external('det64s',
  AA::ARRAY(1..nn,1..nn,integer[8],order=C_order), 
  nn::integer[4], 
  pp::integer[8], 
  LIB=cat(CCodeDir,"detinterp4.so"), 
  RETURN::integer[8]):
Det64s := subsop(1=(AA,nn,p),eval(Det64s));

DetBareissSymm := define_external('detBareissSymmetric', 
  bb::ARRAY(0..nn,datatype=integer[8]),
  nn::integer[4],
  pp::integer[8],
  LIB=cat(CCodeDir,"detinterp4.so"),
  RETURN::integer[8]):
DetBareissSym := subsop(1=(bb,nn,pp),eval(DetBareissSymm));

evallistlist := define_external('evalLIST64s',
    LL::integer[8],
    XX::integer[8],
    aa::ARRAY(1..nn,integer[8]),
    AA::ARRAY(1..nn,1..nn,integer[8]),
    pp::integer[8],
    RETURN::integer[8],
    LIB=cat(CCodeDir,"gcd8.so") ): 
evallistlist := subsop(1=(LL,XX,aa,AA,pp),eval(evallistlist)):

NewtonInv := define_external('NewtonInverses', 
  aa::ARRAY(0..nn,datatype=integer[8]), #size n, indexed from 0 to n-1
  nn::integer[8],                       #size n, deg n-1
  II::ARRAY(0..nn,datatype=integer[8]), #size n, indexed from 0 to n-1
  pp::integer[8],
  LIB=cat(CCodeDir, "detinterp4.so") ):

Interp2var := define_external('interp2var2',
  vii::integer[4], vjj::integer[4],     # 1 <= xi < xj <= N (Maple index)
  dii::integer[8], djj::integer[8], dmm::integer[8], # dmax = max(di,dj)
  XX::ARRAY(0..nn,datatype=integer[8]), # size dmax+1, where dmax=max(di,dj) 
  II::ARRAY(0..nn,datatype=integer[8]), # size dmax+1
  MM::ARRAY(0..didj,datatype=integer[8]), # size (dj+1)*(di+1)
  pp::integer[8],
  LIB=cat(CCodeDir,"detinterp4.so") ):

evalnext := define_external('evalnext64s',
  cc::ARRAY(1..nn,integer[8]), 
  yy::ARRAY(1..nn,integer[8]), 
  nn::integer[4], 
  pp::integer[8],
  LIB=cat(CCodeDir,"gcd5.so") ):

evaladd := define_external('evaladd64s',
  cc::ARRAY(1..nn,integer[8]), 
  nn::integer[4], 
  TT::ARRAY(1..nn,integer[4]), 
  dx::integer[4],
  rr::ARRAY(0..dx,integer[8]), 
  pp::integer[8],
  RETURN::integer[4],
  LIB=cat(CCodeDir,"gcd5.so") ):

getsupport := define_external('getsupport64s',
  ff::integer[8],
  nf::integer[4],
  ALPHA::ARRAY(1..nn,integer[8]),
  nn::integer[4],
  xx::integer[4],
  dx::integer[4],
  TT::ARRAY(0..nn,integer[4]),
  CC::ARRAY(1..nn,integer[8]),
  MM::ARRAY(1..nn,integer[8]),
  pp::integer[8],
  LIB=cat(CCodeDir,"gcd5.so") ):

BHL_C := define_external('HenselLiftCubic',
  AA::ARRAY(0..dxdz,integer[8]), dx::integer[4], dz::integer[4],
  F0::ARRAY(1..nn,0..du,integer[8],order=C_order), nn::integer[4], du::integer[4], 
  FF::ARRAY(1..nn,0..dxdz,integer[8],order=C_order), alpha::integer[8],
  WW::ARRAY(0..dW,integer[8],order=C_order),
  pp::integer[8], 
  LIB=cat(CCodeDir,"hensel2.so"), 
  RETURN::integer[4]):

EXPANDF := define_external('ExpandF',
  FF::ARRAY(1..nn,0..dxdz,integer[8],order=C_order),
  nn::integer[4],
  dx::integer[8],
  dy::integer[8],
  alpha::integer[8],
  pp::integer[8],
  LIB=cat(CCodeDir,"hensel2.so") ):

pointer := proc(f) local MaxInt := 2^63-1; addressof(f)-4*MaxInt end:

# This version evaluates into the matrix A
EVALMOD1 := proc(L::list,X::list,alpha::Array,A::Array,p::integer)
    evallistlist(pointer(L),pointer(X),alpha,A,p);
end:

BB := proc( AAL::list, X::list, alpha::Array, n::integer, p::prime ) global AAa, CNT, tteval, ttdet;
    # Input: AAL is a list of polynomials (entries) from the matrix AA; X=[x1,...,xN], list of variables; 
    #        alpha=[alpha11,...,alphaN]; AAa is a 2d array (global) of size n by n; p is prime 
    # Ouput: Det(AA(alpha)) mod p
    local d, st, et;  CNT += 1;
    st := time(); 
    EVALMOD1( AAL, X, alpha, AAa, p ); # evaluation results stored onto AAa 
    et := time()-st; tteval := tteval + et;
    st := time();
    d := Det64s(AAa, n, p); # Det is computed in C
    et := time()-st; ttdet := ttdet + et; 
    return d;
end:

BBZ := proc( X::list, alpha::{list,Array}, N::integer ) global AA, CNT; 
    # Input: X=[x_1,...,x_N]; alpha=[alpha_1,...,alpha_N]; N is the number of variables
    # Ouput: Det(A(alpha)) where A is a matrix with polynomial entries
    local evalid, C, i;  CNT += 1;
    evalid := [seq(X[i]=alpha[i],i=1..N)]; 
    C := subs(evalid,AA);
    return LinearAlgebra:-Determinant(C); # Determinant is computed in Maple
end:

degBB := proc( AAL::list, X::list, N::integer, j::posint, n::integer, p::prime ) global CNT;
    # Input: AAL is a list of polynomials (entries) from the matrix AA; X=[x1,...,xN], list of variables;
    # Output: deg(a,x_j) where a = det(AA) (1<=j<=N) (w.h.p.)
    local randzp:=rand(p), beta,i,s,F,g,gp,x;
    if j > N then printf("degBB: variable out of range\n"); return 'FAIL'; fi;
    beta := Array(1..N,i->randzp(),datatype=integer[8]);
    for s from 1 do
        beta[j] := s;
        F[s] := BB( AAL, X, beta, n, p ); 
        g := Interp([seq(i,i=1..s)], [seq(F[i],i=1..s)], x) mod p;
        if g = gp then break fi;
        gp := g;
    od;
    if g = 0 then return -1; else return degree(g,x); fi;
end:

Interp1varZ := proc( X::list, alpha::{list,Array}, N::integer, j::posint, deg::list ) global CNT;
    # Input: X=[x_1,...,x_N], alpha=[alpha_1,...,alpha_N], deg is the list of degrees of all variables
    # Output: Det(AA(alpha_1,...,x_j,...,alpha_N))
    local dj, xx, i, F, beta;
    if nops(X)<>N then printf("Interp1varZ: size of X incorrect\n"); return 'FAIL'; fi;
    if j > N then printf("Interp1varZ: index out of range\n"); return 'FAIL'; fi;
    dj := deg[j]; xx := Array(1..dj+1,i->i); printf("d_%d = %d\n",j,dj); 
    beta := Array(alpha); F := Array(1..dj+1); 
    for i to dj+1 do 
        beta[j] := i; F[i] := BBZ( X, beta, N, p ); 
    od;
    return interp( xx, F, X[j] );
end:

GetEvaluations := proc( AAL::list, X::list, beta::Array, var::list, deg::list, xx::Array, MM::Array, n::integer, p::prime ) global CNT;
    # Input:  X=[X_1,...,X_N],beta=[beta1,...,betaN], var=[vi,vj],deg=[di,dj], xx is the Array of evaluation pts, 
    #         indexed from 0 to dmax=max(di,dj), n is the size of the matrix, p is prime
    # Output: Evaluations onto MM = Array(0..(di+1)*(dj+1)-1,datatype=integer[8])
    local alpha, vi, vj, di, dj, i, j; 
    alpha := Array(beta, datatype=integer[8]); 
    vi := var[1]; vj := var[2]; di := deg[1]; dj := deg[2];
    for i from 0 to di do 
        alpha[vi] := xx[i]; 
        for j from 0 to dj do 
            alpha[vj] := xx[j];
            MM[(di+1)*j+i] := BB( AAL, X, alpha, n, p ); 
        od;
    od;
    return;
end:  

support := proc(f, X::list) local m; 
    coeffs(f, X, 'm'); [m]; 
end:
MonCoffs := proc(f, X::list) local m; 
    return [coeffs(f, X, 'm')], [m];
end:
Tcof := proc(T, i::integer, r::integer)
    Array(op(op(op(op(op(T)[2])[r])[2])[i])[2],datatype=integer[8]);
end:
Tmon := proc(T, i::integer, r::integer)
    op(op(op(op(op(T)[2])[r])[2])[i])[1];
end:

nextimage := proc(Ca,Ya,na,Ta,da,Aa,x,p)
local dega,f,i;
    evalnext(Ca,Ya,na,p);
    dega := evaladd(Ca,na,Ta,da,Aa,p);
    f := add(Aa[i]*x^i,i=0..dega);
end: 

initialf2matrix := proc( f0::list, x, F0::Array )
    # Input: f0 is a list of factors in Zp[x] s.t. a(y=alpha) = mul(f0[i],i=1..r) mod p, X=[x,y]
    # Ouput: Matrix F0 s.t. row rr is the array of coefficients of f0[rr] for rr=1..r
    local rr,i,r:=nops(f0),df0:=map(degree,f0,x);
    for rr to r do 
        for i from 0 to df0[rr] do F0[rr,i] := coeff(f0[rr],x,i); od; 
    od;
    return;
end:

clearArray := proc( A::Array, index1::integer, index2::integer ) local i;
    for i from index1 to index2 do A[i] := 0; od;
    return; 
end:

clearArray2d := proc( A::Array, indexi1::integer, indexi2::integer, indexj1::integer, indexj2::integer ) local i,j; 
# Clear a 2-d Array A = Array(indexi1..indexi2,indexj1..indexj2)
    for i from indexi1 to indexi2 do 
        for j from indexj1 to indexj2 do A[i,j] := 0; od;
    od;
    return; 
end:

CMSHL_BBstepj := proc( AAL::list, fjm1::list, j::integer, X::list, alpha::Array, deg::list, N::posint, n::integer, p::prime ) global CNT,tteval,ttdet;
    # Input: AAL is a list of polynomials (entries) from the matrix AA; 
    #        fjm1 is a list of factors in Zp[x_1,...,x_{j-1}] s.t. aj(x_j=alpha_j) = prod(fjm1[i],i=1..r), r is the no.of factors,
    #        X=[x_1,...,x_N], alpha=[alpha_1,...,alpha_N], deg=[deg(a,x_1),...,deg(a,x_N)], n is the matrix size, p is prime
    # Output: fj = [fj_1,...,fj_r], a list of factors of aj in [x_1,...,x_j] and fj_i(x_j=alpha_j) = fjm1_i for i=1..r. Or FAIL.
    local MapleMaxint,randzp,alpham,r,ntf,rr,F,sigma,nsigma,degfx1,M,s,S,k,test,beta,Y,ss,betaA,rr1var,betaAA,FA,TA,CA,MA,vars,nvars,b,dmax,di,dj,xx,II,MM;
    local du,fA,f0A,sizeW,Wtemp,a,f0,i,ii,jj,f,nf,Mon,Cof,mon,cof,Tf,TT,TCOF,SS,SA,mu,dx1,id,sizeVS,Mtemp,Qtemp,C,termsf,Termsf,fY,aY;
    local st0:=time(),tt2:=0,tt3:=0,tt4:=0,st,et,bbtime:=0,interptime:=0,expandtime:=0,vstime:=0,texpand,tbb,tinterp,tvs, bbevaltime:=tteval,bbdettime:=ttdet;
    MapleMaxint := 2^63-1; randzp := rand(p); alpham := p - alpha[j]; r := nops(fjm1); ntf := map(nops, fjm1); 
    for rr to r do
        F[rr] := addressof(fjm1[rr])-4*MapleMaxint; 
        sigma[rr] := [coeffs(fjm1[rr], X[1], 'monfx1')];
        nsigma[rr] := nops(sigma[rr]);
        degfx1[rr] := map(degree, [monfx1], X[1]); 
        M[rr] := map(support, sigma[rr], X[2..j-1]); 
        s[rr] := map(nops, M[rr]);
    od;
    for k to 20 do test := 1; 
        beta := [seq(randzp(), i=1..j)]; Y := [seq(X[i]=beta[i], i=2..j-1)];
        for rr to r do
            S[rr] := map( Eval, M[rr], Y ) mod p;
            for i to nsigma[rr] do if nops({op(S[rr][i])}) < s[rr][i] then test := 0; break; fi; od;
        od; 
    until test = 1;
    if k > 20 then printf( "CMSHL step %d: FAIL(1): p is likely too small\n", j ); return 'FAIL'; fi;
    ss := max([seq(s[rr],rr=1..r)]); printf("CMSHL step %d: s = %a\n",j,ss);
    betaA := Array(1..j-1,beta[1..j-1],datatype=integer[8]); 
    rr1var := Array(1..r); 
    for rr to r do
        FA[rr] := Array(0..degfx1[rr][1],datatype=integer[8]);
        TA[rr] := Array(0..degfx1[rr][1],datatype=integer[4]);
        CA[rr] := Array(1..ntf[rr],datatype=integer[8]);
        MA[rr] := Array(1..ntf[rr],datatype=integer[8]);
        vars := indets(fjm1[rr]); nvars := nops(vars);
        if nvars = 1 then rr1var[rr] := 1;  
            for i from 0 to degfx1[rr][1] do TA[rr][i] := 1; FA[rr][i] := coeff(fjm1[rr],X[1],i); od;
            for i from 1 to min(ntf[rr],degfx1[rr][1]) do CA[rr][i] := FA[rr][i]; MA[rr][i] := 1; od;
            f0[rr] := fjm1[rr];
        elif nvars < j-1 then 
            betaAA := Array(1..nvars,datatype=integer[8]); id:=1;
            for i from 1 to j-1 do 
                if member(X[i],vars) = 'true' then betaAA[id] := beta[i]; id+=1; fi; 
            od;
            getsupport(F[rr],ntf[rr],betaAA,nvars,1,degfx1[rr][1],TA[rr],CA[rr],MA[rr],p);
        else getsupport(F[rr],ntf[rr],betaA,j-1,1,degfx1[rr][1],TA[rr],CA[rr],MA[rr],p);
        fi;
    od;
    b := Array(1..N,i->1,datatype=integer[8]); 
	for i from j+1 to N do b[i] := alpha[i]; od; 
    di := deg[1]; dj := deg[j]; dmax := max(di,dj); 
    xx := Array(0..dmax, i->i+1, datatype=integer[8]); 
    II := Array(0..dmax, datatype=integer[8]); 
    NewtonInv( xx, dmax+1, II, p );
    MM := Array(0..(di+1)*(dj+1)-1,datatype=integer[8]);
    du := max(seq(degfx1[rr][1],rr=1..r)); 
    f0A := Array(1..r,0..du,datatype=integer[8],order=C_order);
    fA := Array(1..r,0..(di+1)*(dj+1)-1,datatype=integer[8],order=C_order);
    sizeW := r+r*(r-1)*di+2*(di+1)*(dj+1)+9*(di+1)+4*(di+1)*(dj+1)+2*r*(di+1)*(dj+1)+5*r*(di+1)+32*(dj+1);
    Wtemp := Array(1..sizeW,datatype=integer[8],order=C_order);
    
    Tf := Array(1..r); for rr to r do Tf[rr]:=table(); od;
    for k to ss do
        for i to j do b[i] := beta[i]*b[i] mod p; od;
        tbb := time(); 
        GetEvaluations( AAL, X, b, [1,j], [di,dj], xx, MM, n, p );
        bbtime += time()-tbb;
        tinterp := time();
        Interp2var( 1, j, di, dj, dmax, xx, II, MM, p );
        interptime += time()-tinterp;
        #a := Expand(add(add(MM[(di+1)*jj+ii]*X[1]^ii,ii=0..di)*X[j]^jj,jj=0..dj)) mod p;    
        st := time();
        for rr to r do 
            if rr1var[rr] = 0 then 
                f0[rr] := nextimage(CA[rr],MA[rr],ntf[rr],TA[rr],degfx1[rr][1],FA[rr],X[1],p);
            fi;
            for ii from 0 to degfx1[rr][1] do f0A[rr,ii] := FA[rr][ii]; od; 
        od;
        et := time()-st; tt2 := tt2 + et;
        for ii to r do
            for jj from ii+1 to r do
                if Gcd(f0[ii],f0[jj]) mod p <> 1 then printf( "CMSHL step %d: FAIL(2): unlucky evaluation\n", j ); 
                    return 'FAIL'; 
                fi;
            od;
        od; 
        st := time(); #f := BHL( a, f0, [X[1],X[j]], alpha[j], p ); 
        BHL_C( MM, di, dj, f0A, r, du, fA, alpha[j], Wtemp, p ); 
	    et := time()-st; tt3 := tt3 + et;
        #texpand := time();
        EXPANDF( fA, r, di, dj, alpham, p ); 
        #expandtime += time()-texpand;
        f := [seq(Expand(add(add(fA[rr][(di+1)*jj+ii]*X[1]^ii,ii=0..di)*X[j]^jj,jj=0..dj)) mod p, rr=1..r)]; 
        clearArray( Wtemp,1,sizeW ); clearArray2d( f0A,1,r,0,du ); clearArray2d( fA,1,r,0,(di+1)*(dj+1)-1 ); 
        st := time();
        nf[k] := map( nops, f ); 
        for rr to r do TT := Tf[rr];
            Cof, Mon := MonCoffs( f[rr], [X[1],X[j]] );
            for i to nops(Mon) do
                mon := Mon[i]; cof := Cof[i]; 
                if not assigned(TT[mon]) then TT[mon] := Array(1..ss); TT[mon][k] := cof;
                else TT[mon][k] := cof;
                fi;
            od;
        od;
        et := time()-st; tt4 := tt4 + et;
    od;
    Qtemp := Array(0..ss,datatype=integer[8]);
    for rr to r do SS := S[rr];
        mu[rr] := max( [seq(nf[k][rr],k=1..ss)] ); 
        for i to mu[rr] do
            dx1 := degree(Tmon(Tf,i,rr),X[1]); id := ListTools[SearchAll](dx1,degfx1[rr]); 
            sizeVS := s[rr][id]; TCOF := Tcof(Tf,i,rr);
            SA := Array(SS[id],datatype=integer[8]);
            Mtemp := Array(0..sizeVS, datatype=integer[8]); 
            C := Array(1..sizeVS, datatype=integer[8]); 
            if j = N then printf( "H.L. x[%d]: rr = %d,i = %d, s[%d][%d] = %d\n", j, rr, i, rr, id, sizeVS ); fi;
            tvs := time();
            VSolve( SA, TCOF, sizeVS, C, Mtemp, Qtemp, 1, p );
            vstime += time()-tvs;
            termsf[rr][i] := add(C[k]*M[rr][id][k], k=1..s[rr][id]);
        od;
        Termsf[rr] := Expand(add(termsf[rr][i]*Tmon(Tf,i,rr), i=1..mu[rr])) mod p;
    od;
    printf( "Timings: BB=%f,Interp=%f,eval=%f,BHL=%f,table=%f,VStime=%f\n",bbtime,interptime,tt2,tt3,tt4,vstime );
    # Check if a = f_1...f_r by another evaluation point
    for i to j do b[i] := randzp(); od;  Y := [seq(X[i]=b[i],i=1..j)]; 
    for rr to r do fY[rr] := Eval(Termsf[rr],Y) mod p; od; 
    aY := BB( AAL,X,b,n,p ); CNT := CNT + 1; 
    bbevaltime := tteval - bbevaltime; bbdettime := ttdet - bbdettime; 
    printf( "Timings: BBeval=%f, BBdet=%f\n", bbevaltime,bbdettime);
    et := time()-st0;  printf( "total time for step %d  = %f\n", j, et );
    if ( aY <> modp( mul(fY[k],k=1..r), p ) ) then 
        printf( "CMSHL step %d: FAIL(3): weak SHL assumption likely failed\n", j ); return FAIL; 
    else return [seq(Termsf[rr],rr=1..r)];
    fi;
end:

CMSHL_BB := proc( AAL::list, f1::list, nn::integer, X::list, alpha::Array, deg::list, N::posint, n::integer, p::prime ) global CNT; 
    local r,du,di,dj,dmax,xx,II,MM,f0A,fA,sizeW,Wtemp,f,rr,ii,jj,fnm1; 
    if nn = 2 then r := nops(f1);  
        di := deg[1]; dj := deg[nn]; dmax := max(di,dj); 
        xx := Array(0..dmax, i->i+1, datatype=integer[8]); 
        II := Array(0..dmax, datatype=integer[8]); 
        NewtonInv( xx, dmax+1, II, p ); 
        MM := Array(0..(di+1)*(dj+1)-1,datatype=integer[8]);
        GetEvaluations( AAL, X, alpha, [1,nn], [di,dj], xx, MM, n, p );
        Interp2var( 1,nn,di,dj,dmax,xx,II,MM,p );

        du := max(map(degree,f1,X[1]));
	f0A := Array(1..r,0..du,datatype=integer[8],order=C_order);
        initialf2matrix( f1, X[1], f0A );
	fA := Array(1..r,0..(di+1)*(dj+1)-1,datatype=integer[8],order=C_order);
	sizeW := r+r*(r-1)*di+2*(di+1)*(dj+1)+9*(di+1)+4*(di+1)*(dj+1)+2*r*(di+1)*(dj+1)+5*r*(di+1)+32*(dj+1);
	Wtemp := Array(1..sizeW,datatype=integer[8],order=C_order); 

	BHL_C( MM, di, dj, f0A, r, du, fA, alpha[nn], Wtemp, p ); 
        EXPANDF( fA, r, di, dj, p-alpha[nn], p ); 
	f := [seq(Expand(add(add(fA[rr][(di+1)*jj+ii]*X[1]^ii,ii=0..di)*X[nn]^jj,jj=0..dj)) mod p, rr=1..r)]; 
        clearArray( Wtemp,1,sizeW ); clearArray2d( f0A,1,r,0,du ); clearArray2d( fA,1,r,0,(di+1)*(dj+1)-1 ); 
        return f; 
    fi; 
    fnm1 := CMSHL_BB( AAL, f1, nn-1, X, alpha, deg, N, n, p );  
    return CMSHL_BBstepj( AAL, fnm1, nn, X, alpha, deg, N, n, p );
end:

p := 2^31-1: randzp := rand(p): 
p1 := 100003: randzp1 := rand(p1): 

for N from 5 to 10 do   
  n := 2*N;
  X := [seq(x||i,i=1..N)]:
  alpha := Array([seq(randzp1(),i=1..N)],datatype=integer[8]);

  testex := 4:
  if testex = 1 then # Example 1
    f := x[1]^8+2*x[1]*x[2]^2*x[4]^3*x[5]+4*x[1]*x[2]^2*x[3]^3+3*x[1]*x[2]^2*x[4]*x[5]^2+x[2]^2*x[3]*x[4]+2*x[1]^2*x[2]^3*x[3]*x[4]^4-5:
    g := x[1]^8+3*x[1]^2*x[2]*x[3]*x[4]^2*x[5]+5*x[1]^2*x[2]*x[3]^2*x[4]-3*x[4]^2*x[5]^2+4*x[5]:
    AA := Matrix([[f, 3*x[2]*x[3]],[0, g]]):
  elif testex = 2 then # Example 2: Cyclic Matrix
    for i from 0 to N-1 do r[i]:=ListTools:-Rotate(X,-i); od:
    AA := Matrix([seq(r[i],i=0..N-1)]):
  else # Example 3: Symmetric Toeplitz matrix
    AA := Matrix(N,N):
    for j from 1 to N-1 do 
        for k from j+1 to N do AA[k,j] := X[k-j+1]: AA[j,k] := AA[k,j]: od:
        AA[j,j] := X[1]:
    od:
    AA[N,N] := X[1]:
  fi:

  reader := subs(FILENAME=Test||N, proc() read FILENAME; end);
  reader();
  
  st := time():
  d := LinearAlgebra:-Determinant(AA0,method=minor):
  et := time()-st: printf("Det minor time = %f\n", et);
  printf("no.of terms in d = %d\n", nops(d));

  AAL := convert(AA,list);
  AAa := Array(1..n,1..n,datatype=integer[8],order=C_order):

  # Get a1 and factor over Zp
  CNT := 0: tteval := 0; ttdet := 0; 
  st := time():
  for i to N do deg[i] := degBB(AAL,X,N,i,n,p); od:
  degA := [seq(deg[i],i=1..N)]; 
  et := time()-st: printf("time for degree computation = %f\n", et):
  printf("no. of probes for degree compt = %d\n", CNT):
  printf("total times for eval and det computations in BB for deg computation: BBeval=%f, BBdet=%f\n", tteval,ttdet);

  a1 := Interp1varZ( X, alpha, N, 1, degA ); 
  a1f := factors(a1):
  a1f := convert(factor(a1), list);
  f1 := map( modp, a1f, p ); # list of factors of a1
  r1 := nops(f1);

  # Start Algorithm CMSHL 
  #trace(CMSHL_BB); trace(CMSHL_BBstepj); 
  CNT := 0: tteval := 0; ttdet := 0; 
  st := time():
  fN := CMSHL_BB( AAL, f1, N, X, alpha, degA, N, n, p ):
  et := time()-st: printf("Total time for CMSHL = %f\n", et):
  printf("Total no.of probes for CMSHL = %d\n", CNT):
  printf("Total times for eval and det for CMSHL: BBeval=%f, BBdet=%f\n", tteval,ttdet); 
  fN := map( mods, fN, p ):  # list of recovered factors in N variables 
  nfN := map(nops, fN): 
  printf("no.of terms in factors = %a;\n", nfN);

  saver := subs(FILENAME=factors_N||N, proc() save nfN,fN,FILENAME; end):
  saver(): 

od:

# st := time():
# dd := LinearAlgebra:-Determinant(AA,method=minor):
# et := time()-st: printf("Det minor time in Maple = %f\n", et);
# nops(dd);
# simplify(dd-d);
 
# for i to r1 do divide(dd,op(i,fN)); od;
# print(nops(fN));
# st := time(): 
# factor(dd):
# et := time()-st: printf("factorizatin time in Maple = %f\n", et);

